@@ -2,21 +2,23 @@
  *      fm-file-info-job.c
  *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2013-2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of the Libfm library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -24,7 +26,7 @@
  * @short_description: Job to gather information about files.
  * @title: FmFileInfoJob
  *
- * @include: libfm/fm-file-info-job.h
+ * @include: libfm/fm.h
  *
  * The #FmFileInfoJob can be used to get filled #FmFileInfo for some files.
  */
@@ -42,7 +44,14 @@
 static void fm_file_info_job_dispose              (GObject *object);
 static gboolean fm_file_info_job_run(FmJob* fmjob);
 
-const char gfile_info_query_attribs[]="standard::*,unix::*,time::*,access::*,id::filesystem";
+const char gfile_info_query_attribs[]="standard::*,unix::*,time::*,access::*,id::filesystem,metadata::emblems";
+
+enum {
+    GOT_INFO,
+    N_SIGNALS
+};
+
+static int signals[N_SIGNALS];
 
 G_DEFINE_TYPE(FmFileInfoJob, fm_file_info_job, FM_TYPE_JOB);
 
@@ -57,6 +66,26 @@ static void fm_file_info_job_class_init(FmFileInfoJobClass *klass)
 
     job_class = FM_JOB_CLASS(klass);
     job_class->run = fm_file_info_job_run;
+
+    /**
+     * FmDirInfoJob::got-info
+     * @job: a job that emitted the signal
+     * @info: #FmFileInfo of file which info is completed
+     *
+     * The #FmDirInfoJob::got-info signal is emitted for every file info
+     * during a job with FM_FILE_INFO_JOB_EMIT_FOR_EACH_FILE flag set.
+     * This signal may be emitted only if info retrieving was successful.
+     *
+     * Since: 1.2.0
+     */
+    signals[GOT_INFO] =
+        g_signal_new("got-info",
+                     G_TYPE_FROM_CLASS(klass),
+                     G_SIGNAL_RUN_LAST,
+                     G_STRUCT_OFFSET(FmFileInfoJobClass, got_info),
+                     NULL, NULL,
+                     g_cclosure_marshal_VOID__POINTER,
+                     G_TYPE_NONE, 1, G_TYPE_POINTER);
 }
 
 
@@ -121,6 +150,52 @@ FmFileInfoJob* fm_file_info_job_new(FmPathList* files_to_query, FmFileInfoJobFla
     return job;
 }
 
+static void _check_native_display_names(FmPath *path)
+{
+    char *path_str, *disp_name;
+
+    if (path == NULL || _fm_path_get_display_name(path) != NULL)
+        return; /* all done */
+    path_str = fm_path_to_str(path);
+    disp_name = g_filename_display_basename(path_str);
+    g_free(path_str);
+    _fm_path_set_display_name(path, disp_name);
+    g_free(disp_name);
+    _check_native_display_names(fm_path_get_parent(path)); /* recursion */
+}
+
+static void _check_gfile_display_names(FmPath *path, GFile *child)
+{
+    GFile *gf;
+    GFileInfo *inf;
+
+    if (path == NULL || _fm_path_get_display_name(path) != NULL)
+        return; /* all done */
+    gf = g_file_get_parent(child);
+    if (gf == NULL) /* file systems such as search:// don't support this */
+        return;
+    inf = g_file_query_info(gf, G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME","
+                                G_FILE_ATTRIBUTE_STANDARD_EDIT_NAME,
+                            G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL, NULL);
+    if (inf != NULL)
+    {
+        const char *dname = g_file_info_get_edit_name(inf);
+        if (!dname || strcmp(dname, "/") == 0)
+            dname = g_file_info_get_display_name(inf);
+        _fm_path_set_display_name(path, dname);
+        g_object_unref(inf);
+    }
+    _check_gfile_display_names(fm_path_get_parent(path), gf); /* recursion */
+    g_object_unref(gf);
+}
+
+static gpointer _emit_current_file(FmJob* job, gpointer user_data)
+{
+    /* this callback is called from the main thread */
+    g_signal_emit(job, signals[GOT_INFO], 0, user_data);
+    return NULL;
+}
+
 static gboolean fm_file_info_job_run(FmJob* fmjob)
 {
     GList* l;
@@ -156,7 +231,11 @@ static gboolean fm_file_info_job_run(FmJob* fmjob)
 
                 fm_file_info_list_delete_link(job->file_infos, l); /* also calls unref */
             }
+            else if(G_UNLIKELY(job->flags & FM_FILE_INFO_JOB_EMIT_FOR_EACH_FILE))
+                fm_job_call_main_thread(fmjob, _emit_current_file, fi);
             g_free(path_str);
+            /* recursively set display names for path parents */
+            _check_native_display_names(fm_path_get_parent(path));
         }
         else
         {
@@ -178,7 +257,7 @@ static gboolean fm_file_info_job_run(FmJob* fmjob)
                 g_file_info_set_file_type(inf, G_FILE_TYPE_DIRECTORY);
                 g_file_info_set_name(inf, fm_path_get_basename(path));
                 g_file_info_set_display_name(inf, fm_path_get_basename(path));
-                fm_file_info_set_from_gfileinfo(fi, inf);
+                fm_file_info_set_from_g_file_data(fi, gf, inf);
                 g_object_unref(inf);
               }
               else
@@ -193,8 +272,14 @@ static gboolean fm_file_info_job_run(FmJob* fmjob)
                 }
 
                 fm_file_info_list_delete_link(job->file_infos, l); /* also calls unref */
+                goto _next;
               }
             }
+            else if(G_UNLIKELY(job->flags & FM_FILE_INFO_JOB_EMIT_FOR_EACH_FILE))
+                    fm_job_call_main_thread(fmjob, _emit_current_file, fi);
+            /* recursively set display names for path parents */
+            _check_gfile_display_names(fm_path_get_parent(path), gf);
+_next:
             g_object_unref(gf);
         }
         l = next;
