@@ -2,21 +2,25 @@
  *      fm-file-ops-xfer.c
  *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2012 Vadim Ushakov <igeekless@gmail.com>
+ *      Copyright 2012-2016 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2018 Nikita Sirgienko <warquark@gmail.com>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of the Libfm library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -29,7 +33,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include "fm-monitor.h"
+#include "fm-utils.h"
 #include <glib/gi18n-lib.h>
 
 static const char query[]=
@@ -48,7 +52,12 @@ static gboolean _fm_file_ops_job_check_paths(FmFileOpsJob* job, GFile* src, GFil
 {
     GError* err = NULL;
     FmJob* fmjob = FM_JOB(job);
-    if(job->type == FM_FILE_OP_MOVE && g_file_equal(src, dest))
+    if(dest == NULL) /* as with search:/// */
+    {
+        err = g_error_new_literal(G_IO_ERROR, G_IO_ERROR_FAILED,
+            _("Destination does not exist"));
+    }
+    else if(job->type == FM_FILE_OP_MOVE && g_file_equal(src, dest))
     {
         err = g_error_new_literal(G_IO_ERROR, G_IO_ERROR_FAILED,
             _("Source and destination are the same."));
@@ -77,9 +86,11 @@ static gboolean _fm_file_ops_job_check_paths(FmFileOpsJob* job, GFile* src, GFil
     return (err == NULL);
 }
 
-static gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
+static gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src,
+                                           GFileInfo* inf, GFile* dest,
+                                           FmFolder *src_folder, /* if move */
+                                           FmFolder *dest_folder)
 {
-    /* FIXME: prevent copying to self or copying parent dir to child. */
     gboolean ret = FALSE;
     gboolean delete_src = FALSE;
     GError* err = NULL;
@@ -88,12 +99,11 @@ static gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileI
     GFile* new_dest = NULL;
     GFileCopyFlags flags;
     FmJob* fmjob = FM_JOB(job);
+    FmPath *fm_dest;
     guint32 mode;
     gboolean skip_dir_content = FALSE;
 
-    /* FIXME: g_file_get_child() failed? generate error! */
-    g_return_val_if_fail(dest != NULL, FALSE);
-
+    job->supported_options = FM_FILE_OP_RENAME | FM_FILE_OP_SKIP | FM_FILE_OP_OVERWRITE;
     if( G_LIKELY(inf) )
         g_object_ref(inf);
     else
@@ -111,7 +121,7 @@ _retry_query_src_info:
         }
     }
 
-    if(!_fm_file_ops_job_check_paths(job, src, inf, dest))
+    if(!_fm_file_ops_job_check_paths(job, src, inf, dest)) /* also checks dest */
     {
         g_object_unref(inf);
         return FALSE;
@@ -142,15 +152,18 @@ _retry_mkdir:
             if( !fm_job_is_cancelled(fmjob) && !job->skip_dir_content &&
                 !g_file_make_directory(dest, fm_job_get_cancellable(fmjob), &err) )
             {
-                if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
+                if(err->domain == G_IO_ERROR && (err->code == G_IO_ERROR_EXISTS ||
+                                                 err->code == G_IO_ERROR_INVALID_FILENAME ||
+                                                 err->code == G_IO_ERROR_FILENAME_TOO_LONG))
                 {
                     GFile* dest_cp = new_dest;
+                    gboolean dest_exists = (err->code == G_IO_ERROR_EXISTS);
                     FmFileOpOption opt = 0;
                     g_error_free(err);
                     err = NULL;
 
                     new_dest = NULL;
-                    opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                    opt = _fm_file_ops_job_ask_new_name(job, src, dest, &new_dest, dest_exists);
                     if(!new_dest) /* restoring status quo */
                         new_dest = dest_cp;
                     else if(dest_cp) /* we got new new_dest, forget old one */
@@ -213,9 +226,6 @@ _retry_chmod_for_dir:
                 }
                 job->finished += size;
                 fm_file_ops_job_emit_percent(job);
-
-                if(job->dest_folder_mon)
-                    g_file_monitor_emit_event(job->dest_folder_mon, dest, NULL, G_FILE_MONITOR_EVENT_CREATED);
             }
 
             if(!dir_created) /* if target dir is not created, don't copy dir content */
@@ -228,6 +238,20 @@ _retry_chmod_for_dir:
             /* FIXME: handle the case when the dir cannot be created. */
             else if(!fm_job_is_cancelled(fmjob))
             {
+                FmFolder *sub_folder;
+                FmFolder *sub_src = NULL;
+
+                if (delete_src)
+                {
+                    FmPath *src_path = fm_path_new_for_gfile(src);
+                    sub_src = fm_folder_find_by_path(src_path);
+                    fm_path_unref(src_path);
+                }
+                fm_dest = fm_path_new_for_gfile(dest);
+                sub_folder = fm_folder_find_by_path(fm_dest);
+                /* inform folder we created directory */
+                if (!dest_folder || !_fm_folder_event_file_added(dest_folder, fm_dest))
+                    fm_path_unref(fm_dest);
 _retry_enum_children:
                 enu = g_file_enumerate_children(src, query, G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
                                             fm_job_get_cancellable(fmjob), &err);
@@ -252,7 +276,6 @@ _retry_enum_children:
                             else
                             {
                                 gboolean ret2;
-                                GFileMonitor* old_dest_mon = job->dest_folder_mon;
                                 GFile* sub = g_file_get_child(src, g_file_info_get_name(inf));
                                 GFile* sub_dest;
                                 char* tmp_basename;
@@ -261,26 +284,21 @@ _retry_enum_children:
                                     /* both are native or both are virtual */
                                     tmp_basename = NULL;
                                 else if(g_file_is_native(src)) /* copy from native to virtual */
-                                    tmp_basename = NULL; /* gvfs escapes it itself */
+                                    tmp_basename = g_filename_to_utf8(g_file_info_get_name(inf),
+                                                                      -1, NULL, NULL, NULL);
+                                    /* gvfs escapes it itself */
                                 else /* copy from virtual to native */
-                                    tmp_basename = g_uri_unescape_string(g_file_info_get_name(inf), NULL);
+                                    /* display name as copying target more prefers, that basename (for example, for Google Drive)
+                                    See https://debarshiray.wordpress.com/2015/09/13/google-drive-and-gnome-what-is-a-volatile-path/ for some explanations*/
+                                    tmp_basename = fm_uri_subpath_to_native_subpath(g_file_info_get_display_name(inf), NULL);
                                 sub_dest = g_file_get_child(dest,
                                         tmp_basename ? tmp_basename : g_file_info_get_name(inf));
                                 g_free(tmp_basename);
 
-                                if(g_file_is_native(dest))
-                                    job->dest_folder_mon = NULL;
-                                else
-                                    job->dest_folder_mon = fm_monitor_lookup_dummy_monitor(dest);
-
-                                ret2 = _fm_file_ops_job_copy_file(job, sub, inf, sub_dest);
+                                ret2 = _fm_file_ops_job_copy_file(job, sub, inf, sub_dest, sub_src, sub_folder);
                                 g_object_unref(sub);
                                 g_object_unref(sub_dest);
 
-                                if(job->dest_folder_mon)
-                                    g_object_unref(job->dest_folder_mon);
-                                job->dest_folder_mon = old_dest_mon;
-
                                 if(ret2)
                                     ++n_copied;
                                 else
@@ -292,7 +310,6 @@ _retry_enum_children:
                         {
                             if(err)
                             {
-                                /* FIXME: error handling */
                                 fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
                                 g_error_free(err);
                                 err = NULL;
@@ -333,6 +350,10 @@ _retry_enum_children:
                     if(act == FM_JOB_RETRY)
                         goto _retry_enum_children;
                 }
+                if (sub_src)
+                    g_object_unref(sub_src);
+                if (sub_folder)
+                    g_object_unref(sub_folder);
             }
             if(job->skip_dir_content)
                 delete_src = FALSE;
@@ -359,24 +380,20 @@ _retry_enum_children:
                     int r = mkfifo(dest_path, src_st.st_mode);
                     g_free(dest_path);
                     if( r == 0)
-                    {
                         ret = TRUE;
-                        break;
-                    }
-                    else
-                    {
-                        /* FIXME: error handling */
-                    }
                 }
                 /* FIXME: how about block device, char device, and socket? */
             }
-            else
-            {
-                /* FIXME: error handling */
-            }
         }
-        job->finished += size;
-        fm_file_ops_job_emit_percent(job);
+        if (!ret)
+        {
+            g_set_error(&err, G_IO_ERROR, G_IO_ERROR_FAILED,
+                        _("Cannot copy file '%s': not supported"),
+                        g_file_info_get_display_name(inf));
+            fm_job_emit_error(fmjob, err, FM_JOB_ERROR_MODERATE);
+            g_clear_error(&err);
+        }
+        goto _file_copied;
 
     default:
         flags = G_FILE_COPY_ALL_METADATA|G_FILE_COPY_NOFOLLOW_SYMLINKS;
@@ -386,16 +403,19 @@ _retry_copy:
         {
             flags &= ~G_FILE_COPY_OVERWRITE;
 
-            /* handle existing files */
-            if(err->domain == G_IO_ERROR && err->code == G_IO_ERROR_EXISTS)
+            /* handle existing files or file name conflict */
+            if(err->domain == G_IO_ERROR && (err->code == G_IO_ERROR_EXISTS ||
+                                             err->code == G_IO_ERROR_INVALID_FILENAME ||
+                                             err->code == G_IO_ERROR_FILENAME_TOO_LONG))
             {
                 GFile* dest_cp = new_dest;
+                gboolean dest_exists = (err->code == G_IO_ERROR_EXISTS);
                 FmFileOpOption opt = 0;
                 g_error_free(err);
                 err = NULL;
 
                 new_dest = NULL;
-                opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                opt = _fm_file_ops_job_ask_new_name(job, src, dest, &new_dest, dest_exists);
                 if(!new_dest) /* restoring status quo */
                     new_dest = dest_cp;
                 else if(dest_cp) /* we got new new_dest, forget old one */
@@ -442,13 +462,15 @@ _retry_copy:
         else
             ret = TRUE;
 
+_file_copied:
         job->finished += size;
         job->current_file_finished = 0;
 
-        if( ret && type != G_FILE_TYPE_DIRECTORY )
+        if(ret && dest_folder)
         {
-            if(job->dest_folder_mon)
-                g_file_monitor_emit_event(job->dest_folder_mon, dest, NULL, G_FILE_MONITOR_EVENT_CREATED);
+            fm_dest = fm_path_new_for_gfile(dest);
+            if(!_fm_folder_event_file_added(dest_folder, fm_dest))
+                fm_path_unref(fm_dest);
         }
 
         /* update progress */
@@ -458,7 +480,7 @@ _retry_copy:
     /* if this is a cross-device move operation, delete source files. */
     /* ret == TRUE means the copy is successful. */
     if( !fm_job_is_cancelled(fmjob) && ret && delete_src )
-        ret = _fm_file_ops_job_delete_file(fmjob, src, inf); /* delete the source file. */
+        ret = _fm_file_ops_job_delete_file(fmjob, src, inf, src_folder, TRUE); /* delete the source file. */
 
     if(new_dest)
         g_object_unref(new_dest);
@@ -466,15 +488,17 @@ _retry_copy:
     return ret;
 }
 
-gboolean _fm_file_ops_job_move_file(FmFileOpsJob* job, GFile* src, GFileInfo* inf, GFile* dest)
+gboolean _fm_file_ops_job_move_file(FmFileOpsJob* job, GFile* src,
+                                    GFileInfo* inf, GFile* dest, FmPath *src_path,
+                                    FmFolder *src_folder, FmFolder *dest_folder)
 {
-    /* FIXME: prevent moving to self or moving parent dir to child. */
     GError* err = NULL;
     FmJob* fmjob = FM_JOB(job);
     const char* src_fs_id;
     gboolean ret = TRUE;
     GFile* new_dest = NULL;
 
+    job->supported_options = FM_FILE_OP_RENAME | FM_FILE_OP_SKIP | FM_FILE_OP_OVERWRITE;
     if( G_LIKELY(inf) )
         g_object_ref(inf);
     else
@@ -504,7 +528,9 @@ _retry_query_src_info:
     {
         guint64 size;
         GFileCopyFlags flags = G_FILE_COPY_ALL_METADATA|G_FILE_COPY_NOFOLLOW_SYMLINKS;
+        FmPath *fm_dest;
 
+        fm_dest = fm_path_new_for_gfile(dest);
         /* showing currently processed file. */
         fm_file_ops_job_emit_cur_file(job, g_file_info_get_display_name(inf));
 _retry_move:
@@ -517,7 +543,7 @@ _retry_move:
                 FmFileOpOption opt = 0;
 
                 new_dest = NULL;
-                opt = fm_file_ops_job_ask_rename(job, src, NULL, dest, &new_dest);
+                opt = _fm_file_ops_job_ask_new_name(job, src, dest, &new_dest, TRUE);
                 if(!new_dest) /* restoring status quo */
                     new_dest = dest_cp;
                 else if(dest_cp) /* we got new new_dest, forget old one */
@@ -541,6 +567,8 @@ _retry_move:
                         if(enu)
                         {
                             GFileInfo* child_inf;
+                            FmFolder *sub_src_folder = fm_folder_find_by_path(src_path);
+                            FmFolder *sub_dest_folder = fm_folder_find_by_path(fm_dest);
                             while(!fm_job_is_cancelled(fmjob))
                             {
                                 child_inf = g_file_enumerator_next_file(enu, fm_job_get_cancellable(fmjob), &err);
@@ -548,10 +576,12 @@ _retry_move:
                                 {
                                     GFile* child = g_file_get_child(src, g_file_info_get_name(child_inf));
                                     GFile* child_dest = g_file_get_child(dest, g_file_info_get_name(child_inf));
-                                    _fm_file_ops_job_move_file(job, child, child_inf, child_dest);
+                                    FmPath *child_path = fm_path_new_for_gfile(child);
+                                    _fm_file_ops_job_move_file(job, child, child_inf, child_dest, child_path, sub_src_folder, sub_dest_folder);
                                     g_object_unref(child);
                                     g_object_unref(child_dest);
                                     g_object_unref(child_inf);
+                                    fm_path_unref(child_path);
                                 }
                                 else
                                 {
@@ -567,6 +597,10 @@ _retry_move:
                                     }
                                 }
                             }
+                            if (sub_src_folder)
+                                g_object_unref(sub_src_folder);
+                            if (sub_dest_folder)
+                                g_object_unref(sub_dest_folder);
                             g_object_unref(enu);
                         }
                         else
@@ -609,13 +643,14 @@ _retry_move:
                 if(act == FM_JOB_RETRY)
                     goto _retry_move;
             }
+            fm_path_unref(fm_dest);
         }
         else
         {
-            if(job->src_folder_mon)
-                g_file_monitor_emit_event(job->src_folder_mon, src, NULL, G_FILE_MONITOR_EVENT_DELETED);
-            if(job->dest_folder_mon)
-                g_file_monitor_emit_event(job->dest_folder_mon, dest, NULL, G_FILE_MONITOR_EVENT_CREATED);
+            if (src_folder)
+                _fm_folder_event_file_deleted(src_folder, src_path);
+            if (!dest_folder || !_fm_folder_event_file_added(dest_folder, fm_dest))
+                fm_path_unref(fm_dest);
         }
 /*
         size = g_file_info_get_attribute_uint64(inf, G_FILE_ATTRIBUTE_UNIX_BLOCKS);
@@ -630,7 +665,7 @@ _retry_move:
     {
         /* use copy & delete */
         /* source file will be deleted in _fm_file_ops_job_copy_file() */
-        ret = _fm_file_ops_job_copy_file(job, src, inf, dest);
+        ret = _fm_file_ops_job_copy_file(job, src, inf, dest, src_folder, dest_folder);
     }
 
     if(new_dest)
@@ -652,14 +687,14 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
 {
     gboolean ret = TRUE;
     GFile *dest_dir;
-    GFileMonitor *old_mon, *dest_mon;
     GList* l;
     FmJob* fmjob = FM_JOB(job);
     /* prepare the job, count total work needed with FmDeepCountJob */
     FmDeepCountJob* dc = fm_deep_count_job_new(job->srcs, FM_DC_JOB_DEFAULT);
+    FmFolder *df;
+
     /* let the deep count job share the same cancellable object. */
     fm_job_set_cancellable(FM_JOB(dc), fm_job_get_cancellable(fmjob));
-    /* FIXME: there is no way to cancel the deep count job here. */
     fm_job_run_sync(FM_JOB(dc));
     job->total = dc->total_size;
     if(fm_job_is_cancelled(fmjob))
@@ -671,18 +706,10 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
     g_debug("total size to copy: %llu", (long long unsigned int)job->total);
 
     dest_dir = fm_path_to_gfile(job->dest);
-    /* get dummy file monitors for non-native filesystems */
-    if( g_file_is_native(dest_dir) )
-    {
-        old_mon = NULL; /* to satisfy compiler */
-        dest_mon = NULL;
-    }
-    else
-    {
-        old_mon = job->dest_folder_mon;
-        dest_mon = fm_monitor_lookup_dummy_monitor(dest_dir);
-        job->dest_folder_mon = dest_mon;
-    }
+    /* suspend updates for destination */
+    df = fm_folder_find_by_path(job->dest);
+    if (df)
+        fm_folder_block_updates(df);
 
     fm_file_ops_job_emit_prepared(job);
 
@@ -693,38 +720,61 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
         GFile* dest;
         char* tmp_basename;
 
-        if(g_file_is_native(src) == g_file_is_native(dest_dir))
-            /* both are native or both are virtual */
+        if(g_file_is_native(src) && g_file_is_native(dest_dir))
+            /* both are native */
             tmp_basename = NULL;
         else if(g_file_is_native(src)) /* copy from native to virtual */
-            tmp_basename = NULL; /* gvfs escapes it itself */
-        else /* copy from virtual to native */
-            tmp_basename = g_uri_unescape_string(fm_path_get_basename(path), NULL);
+            tmp_basename = g_filename_to_utf8(fm_path_get_basename(path),
+                                              -1, NULL, NULL, NULL);
+            /* gvfs escapes it itself */
+        else /* copy from virtual to native/virtual */
+        {
+            /* if we drop URI query onto native filesystem, omit query part */
+            const char *basename = fm_path_get_basename(path);
+            char *sub_name;
+
+            sub_name = strchr(basename, '?');
+            if (sub_name)
+            {
+                sub_name = g_strndup(basename, sub_name - basename);
+                basename = strrchr(sub_name, G_DIR_SEPARATOR);
+                if (basename)
+                    basename++;
+                else
+                    basename = sub_name;
+                tmp_basename = fm_uri_subpath_to_native_subpath(basename, NULL);
+                g_free(sub_name);
+            }
+            else
+                /* if not URI, better use display name */
+                tmp_basename = fm_path_display_basename(path);
+        }
         dest = g_file_get_child(dest_dir,
                         tmp_basename ? tmp_basename : fm_path_get_basename(path));
         g_free(tmp_basename);
-        if(!_fm_file_ops_job_copy_file(job, src, NULL, dest))
+        if(!_fm_file_ops_job_copy_file(job, src, NULL, dest, NULL, df))
             ret = FALSE;
         g_object_unref(src);
-        g_object_unref(dest);
+        if(dest != NULL)
+            g_object_unref(dest);
     }
 
     /* g_debug("finished: %llu, total: %llu", job->finished, job->total); */
     fm_file_ops_job_emit_percent(job);
 
-    g_object_unref(dest_dir);
-    if(dest_mon)
+    /* restore updates for destination */
+    if (df)
     {
-        g_object_unref(dest_mon);
-        job->dest_folder_mon = old_mon;
+        fm_folder_unblock_updates(df);
+        g_object_unref(df);
     }
+    g_object_unref(dest_dir);
     return ret;
 }
 
 gboolean _fm_file_ops_job_move_run(FmFileOpsJob* job)
 {
     GFile *dest_dir;
-    GFileMonitor *old_mon, *dest_mon, *old_src_mon;
     GFileInfo* inf;
     GList* l;
     GError* err = NULL;
@@ -732,6 +782,8 @@ gboolean _fm_file_ops_job_move_run(FmFileOpsJob* job)
     dev_t dest_dev = 0;
     gboolean ret = TRUE;
     FmDeepCountJob* dc;
+    FmPath *parent = NULL;
+    FmFolder *df, *sf = NULL;
 
     /* get information of destination folder */
     g_return_val_if_fail(job->dest, FALSE);
@@ -778,22 +830,12 @@ _retry_query_dest_info:
     g_debug("total size to move: %llu, dest_fs: %s",
             (long long unsigned int)job->total, job->dest_fs_id);
 
-    /* get dummy file monitors for non-native filesystems */
-    if( g_file_is_native(dest_dir) )
-    {
-        old_mon = NULL; /* to satisfy compiler */
-        dest_mon = NULL;
-    }
-    else
-    {
-        old_mon = job->dest_folder_mon;
-        dest_mon = fm_monitor_lookup_dummy_monitor(dest_dir);
-        job->dest_folder_mon = dest_mon;
-    }
-
     fm_file_ops_job_emit_prepared(job);
+    /* suspend updates for destination */
+    df = fm_folder_find_by_path(job->dest);
+    if (df)
+        fm_folder_block_updates(df);
 
-    old_src_mon = job->src_folder_mon;
     for(l = fm_path_list_peek_head_link(job->srcs); !fm_job_is_cancelled(fmjob) && l; l=l->next)
     {
         FmPath* path = FM_PATH(l->data);
@@ -801,46 +843,65 @@ _retry_query_dest_info:
         GFile* dest;
         char* tmp_basename;
 
-        if(g_file_is_native(src) == g_file_is_native(dest_dir))
-            /* both are native or both are virtual */
+        /* do with updates for source */
+        if (fm_path_get_parent(path) != parent && fm_path_get_parent(path) != NULL)
+        {
+            FmFolder *pf;
+
+            pf = fm_folder_find_by_path(fm_path_get_parent(path));
+            if (pf != sf)
+            {
+                if (sf)
+                {
+                    fm_folder_unblock_updates(sf);
+                    g_object_unref(sf);
+                }
+                if (pf)
+                    fm_folder_block_updates(pf);
+                sf = pf;
+            }
+            else if (pf)
+                g_object_unref(pf);
+        }
+        parent = fm_path_get_parent(path);
+        if(g_file_is_native(src) && g_file_is_native(dest_dir))
+            /* both are native */
             tmp_basename = NULL;
         else if(g_file_is_native(src)) /* move from native to virtual */
-            tmp_basename = NULL; /* gvfs escapes it itself */
-        else /* move from virtual to native */
-            tmp_basename = g_uri_unescape_string(fm_path_get_basename(path), NULL);
+            tmp_basename = g_filename_to_utf8(fm_path_get_basename(path),
+                                              -1, NULL, NULL, NULL);
+            /* gvfs escapes it itself */
+        else /* move from virtual to native/virtual */
+        {
+            char *name = fm_path_display_basename(path);
+            tmp_basename = fm_uri_subpath_to_native_subpath(name, NULL);
+            g_free(name);
+        }
         dest = g_file_get_child(dest_dir,
                         tmp_basename ? tmp_basename : fm_path_get_basename(path));
         g_free(tmp_basename);
-        /* get dummy file monitors for non-native filesystems */
-        job->src_folder_mon = NULL;
-        if(!g_file_is_native(src))
-        {
-            GFile* src_dir = g_file_get_parent(src);
-            if(src_dir)
-            {
-                job->src_folder_mon = fm_monitor_lookup_dummy_monitor(src_dir);
-                g_object_unref(src_dir);
-            }
-        }
 
-        if(!_fm_file_ops_job_move_file(job, src, NULL, dest))
+        if(!_fm_file_ops_job_move_file(job, src, NULL, dest, path, sf, df))
             ret = FALSE;
         g_object_unref(src);
-        g_object_unref(dest);
-
-        if(job->src_folder_mon)
-            g_object_unref(job->src_folder_mon);
+        if(dest != NULL)
+            g_object_unref(dest);
 
         if(!ret)
             break;
     }
-    job->src_folder_mon = old_src_mon;
+    /* restore updates for destination and source */
+    if (df)
+    {
+        fm_folder_unblock_updates(df);
+        g_object_unref(df);
+    }
+    if (sf)
+    {
+        fm_folder_unblock_updates(sf);
+        g_object_unref(sf);
+    }
 
     g_object_unref(dest_dir);
-    if(dest_mon)
-    {
-        g_object_unref(dest_mon);
-        job->dest_folder_mon = old_mon;
-    }
     return ret;
 }
